/*
								+----------------------------------+
								|                                  |
								| *** Memory consistency bench *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

/*
 *	2008-08-19
 *
 *	this defines __USE_MALLOC to avoid stl templates leaving memory allocated at the end
 *	also newly using both low-level and high-level shaders
 *
 */

#if defined(WIN32) || defined(_WIN64)
#include <crtdbg.h>
#undef _ASSERTE
#endif
// win32 memory debugger

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#define GL_GLEXT_LEGACY
#include <GL/gl.h>
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <algorithm>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Tga.h"
#include "../berLame_src/Vector.h"

static CGLState *p_state = 0; // OpenGL state guard

bool SaveTexture(const char *p_s_filename, const CGLTexture_2D *p_texture)
{
	TBmp t_bitmap = {8, false, true, p_texture->n_Width(), p_texture->n_Height()};
	if(!(t_bitmap.p_buffer = new uint32_t[p_texture->n_Width() *
	   p_texture->n_Height()])) {
		fprintf(stderr, "error: not enough memory\n");
		return false;
	}
	p_texture->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_BGRA, GL_UNSIGNED_BYTE, t_bitmap.p_buffer);
	// alloc and download bitmap

	bool b_result = CTgaCodec::Save_TGA(p_s_filename, t_bitmap);
	// save to file

	delete[] t_bitmap.p_buffer;
	// cleanup

	return b_result;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(1, 1);
	glutCreateWindow("berLame OpenGL objects memory consistency benchmark");
	// init OpenGL using GLUT

	printf("create state guard\n");

	if(!(p_state = new CGLState())) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create OpenGL state guard

	printf("create empty texture\n");

	CGLTexture_2D *p_texture;
	if(!(p_texture = new CGLTexture_2D(p_state, 256, 256, GL_RGBA8, true, 0))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create 2D texture with unspecified contents

	CGLTexture_2D *p_texture2;
	if(!(p_texture2 = new CGLTexture_2D(p_state, 256, 256, GL_RGBA8, true, 0))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create 2D texture with unspecified contents

	if(!SaveTexture("0_empty.tga", p_texture)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save texture to file

	if(!SaveTexture("1_empty.tga", p_texture2)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save texture to file

	printf("create noise texture\n");

	unsigned char *p_image_data;
	if(!(p_image_data = new unsigned char[3 * 256 * 256])) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	for(unsigned char *p_dest = p_image_data,
	   *p_end = p_image_data + 256 * 256 * 3; p_dest != p_end;)
		*p_dest ++ = rand();
	CGLTexture_2D *p_noise_texture;
	if(!(p_noise_texture = new CGLTexture_2D(p_state, 256, 256,
	   GL_RGBA8, true, 0, GL_RGB, GL_UNSIGNED_BYTE, p_image_data))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	delete[] p_image_data;
	// create 2D texture with random contents

	if(!SaveTexture("2_noise.tga", p_noise_texture)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save texture to file

	printf("create render-buffer for r2t (with depth buffer)\n");

	CGLFrameBuffer *p_frame_buffer;
	if(!CGLFrameBuffer::b_Supported()) {
		fprintf(stderr, "error: CGLFrameBuffer not supported\n");
		return -1;
	}

	if(!(p_frame_buffer = new CGLFrameBuffer(256, 256, 1, true, GL_RGBA8,
	   true, false, GL_DEPTH_COMPONENT24, false, false, 0))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}

	printf("render something to the texture\n");

	{
		if(!p_frame_buffer->b_Status() ||
		   !p_frame_buffer->Bind() ||
		   !p_frame_buffer->BindColorTexture_2D(*p_noise_texture, GL_TEXTURE_2D)) {
			fprintf(stderr, "error: failed to bind FBO for r2t\n");
			return -1;
		}
		glViewport(0, 0, p_frame_buffer->n_Width(), p_frame_buffer->n_Height());
		// bind FBO for rendering to texture

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		// ortho

		p_state->EnableDepthTest();
		glClear(GL_DEPTH_BUFFER_BIT);
		// clear color & depth

		p_state->LineWidth(3);
		p_state->DisableTexture2D();

		p_state->Color3f(.75f, .75f, .75f);
		glBegin(GL_POLYGON);
		for(int i = 0; i < 360; i += 15) {
			float f_angle = i / 180.0f * f_pi;
			glVertex2f(sin(f_angle) * .95f, cos(f_angle) * .95f);
		}
		glEnd();
		// draw polygon

		glTranslatef(0, 0, -.5f);
		p_state->DepthMask(0);
		// prepare for drawing transparent ticks

		p_state->Enable(GL_LINE_SMOOTH);
		p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		p_state->EnableBlend();
		// enable smooth lines

		glBegin(GL_LINES);
		p_state->Color3f(0, 1, 0);
		for(int i = 0; i < 12; ++ i) {
			float f_angle = i / 6.0f * f_pi;
			glVertex2f(sin(f_angle) * .9f, cos(f_angle) * .9f);
			glVertex2f(sin(f_angle) * .7f, cos(f_angle) * .7f);
		}
		p_state->Color3f(0, 0, 0);

		time_t t_time = time(NULL);
		tm *p_time = localtime(&t_time);
		int n_hours = p_time->tm_hour;
		int n_minutes = p_time->tm_min;
		int n_seconds = p_time->tm_sec;
		// get actual time

		float f_minutes = (n_minutes + n_seconds / 60.0f) / 60 * 2 * f_pi;
		float f_hours = (n_hours + (n_minutes + n_seconds / 60.0f) / 60.0f) / 12 * 2 * f_pi;
		// calculate angle in radians

		glVertex2f(0, 0);
		glVertex2f(sin(f_hours) * .3f, -cos(f_hours) * .3f);
		glVertex2f(0, 0);
		glVertex2f(sin(f_minutes) * .8f, -cos(f_minutes) * .8f);
		glEnd();
		p_state->Color3f(1, 0, 0);
		glBegin(GL_LINE_LOOP);
		for(int i = 0; i < 360; i += 5) {
			float f_angle = i / 180.0f * f_pi;
			glVertex2f(sin(f_angle) * .95f, cos(f_angle) * .95f);
		}
		glEnd();
		p_state->DisableBlend();
		p_state->DepthMask(1);
		// draw something

		if(!p_frame_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D) ||
		   !p_frame_buffer->Release()) {
			fprintf(stderr, "error: failed to release FBO\n");
			return -1;
		}
		// release FBO

		p_noise_texture->Bind_Enable(p_state);
		glGenerateMipmapEXT(GL_TEXTURE_2D);
	}
	
	if(!SaveTexture("3_clock.tga", p_noise_texture)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save output to file

	printf("create render-buffer for multi r2t (with depth buffer)\n");

	CGLFrameBuffer *p_frame_buffer2;
	if(!(p_frame_buffer2 = new CGLFrameBuffer(256, 256, 2, true, GL_RGBA8,
	   true, false, GL_DEPTH_COMPONENT24, false, false, 0))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}

	printf("bulid very simple shader, using CGLShader\n");

	CGLShader *p_shader;
	if(!CGL_Core_shader::b_Supported()) {
		fprintf(stderr, "error: CGL_Core_shader not supported\n");
		return -1;
	}

	{
		const char *p_s_red_green = "void main() { gl_FragData[0] ="
			" vec4(1, 0, 0, 1); gl_FragData[1] = vec4(0, 1, 0, 1); }";
		// with intentional warning

		if(!(p_shader = new CGL_Core_shader(0, p_s_red_green))) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		char *p_s_info_log = 0;
		if(!p_shader->Compile(p_state, p_s_info_log)) {
			fprintf(stderr, "error: error compiling shader:\n%s\n",
				(p_s_info_log)? p_s_info_log : "(null)");
			if(p_s_info_log)
				delete[] p_s_info_log;
			return -1;
		}
		if(p_s_info_log) {
			fprintf(stderr, "warning: while compiling shader:\n%s\n", p_s_info_log);
			delete[] p_s_info_log;
			p_s_info_log = 0;
		}
		if(!p_shader->Link(p_s_info_log)) {
			fprintf(stderr, "error: error linking shader:\n%s\n",
				(p_s_info_log)? p_s_info_log : "(null)");
			if(p_s_info_log)
				delete[] p_s_info_log;
			return -1;
		}
		if(p_s_info_log) {
			fprintf(stderr, "warning: while linking shader:\n%s\n", p_s_info_log);
			delete[] p_s_info_log;
			p_s_info_log = 0;
		}
		// compile and link the shader
	}
	// create shader

	{
		if(!p_frame_buffer2->b_Status() ||
		   !p_frame_buffer2->Bind() ||
		   !p_frame_buffer2->BindColorTexture_2D(*p_texture, GL_TEXTURE_2D, 0) ||
		   !p_frame_buffer2->BindColorTexture_2D(*p_texture2, GL_TEXTURE_2D, 1)) {
			fprintf(stderr, "error: failed to bind FBO for r2t\n");
			return -1;
		}
		glViewport(0, 0, p_frame_buffer->n_Width(), p_frame_buffer->n_Height());
		// bind FBO for rendering to texture

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		// ortho

		p_state->ClearColor4f(1, 1, 1, 1);
		p_state->EnableDepthTest();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		// clear color & depth

		p_shader->Bind(p_state);
		// bind shader

		glBegin(GL_QUADS);
		glVertex2f(-1, -1);
		glVertex2f( 1, -1);
		glVertex2f( 1,  1);
		glVertex2f(-1,  1);
		glEnd();
		// draw fullscreen quad

		if(!p_frame_buffer2->BindColorTexture_2D(0, GL_TEXTURE_2D, 0) ||
		   !p_frame_buffer2->BindColorTexture_2D(0, GL_TEXTURE_2D, 1) ||
		   !p_frame_buffer2->Release()) {
			fprintf(stderr, "error: failed to release FBO\n");
			return -1;
		}
		// release FBO

		p_state->BindProgramObject(0);
		// release shader
	}

	if(!SaveTexture("4_left.tga", p_texture2)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save texture to file

	if(!SaveTexture("5_right.tga", p_texture)) {
		fprintf(stderr, "error: error saving texture to file\n");
		return -1;
	}
	// save texture to file

	printf("bulid box filter shader, using CGLShaderObject\n");

	for(int n_level = 0; n_level < 2; ++ n_level) {
		CGLShaderObject *p_shader_object;
		CGLShaderParams *p_shader_params;

		{
			TShaderInfo t_filter;
			if(n_level) {
				t_filter = TShaderInfo("filter", TShaderInfo::proc_Fragment,
					"uniform sampler2D n_tex;\n"
					"uniform vec4 v_pix;\n"
					"void main()\n"
					"{\n"
					"    vec4 v_accum = texture2D(n_tex, gl_TexCoord[0].xy, 1.0);\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy + v_pix.xw, 1.0) * .5;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy - v_pix.xw, 1.0) * .5;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy + v_pix.wy, 1.0) * .5;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy - v_pix.wy, 1.0) * .5;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy + v_pix.xy, 1.0) * .25;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy - v_pix.xy, 1.0) * .25;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy + v_pix.xz, 1.0) * .25;\n"
					"    v_accum += texture2D(n_tex, gl_TexCoord[0].xy - v_pix.xz, 1.0) * .25;\n"
					"    gl_FragColor = v_accum * .25;\n"
					"}\n", "n_tex");
				printf("\tusing high-level OpenGL 2.0 core fragment shader\n");
			} else {
				t_filter = TShaderInfo("filter", TShaderInfo::proc_Fragment,
					"!!ARBfp1.0\n"
					"OPTION NV_fragment_program2;\n"
					"PARAM c[2] = { program.local[0],\n"
					"		{ 1, 0.5, 0.25, 2.0 } };\n"
					"TEMP R0;\n"
					"TEMP R1;\n"
					"TEMP R2;\n"
					"TEMP R3;\n"
					"TEMP RC;\n"
					"TEMP HC;\n"
					"OUTPUT oCol = result.color;\n"
					"ADDR  R3, fragment.texcoord[0].xyxy, c[0].xwwy;\n"
					"ADDR  R2, fragment.texcoord[0].xyxy, -c[0].xwwy;\n"
					"MOVR  R0.z, c[1].x;\n"
					"MOVR  R0.xy, fragment.texcoord[0];\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MOVR  R1.z, c[1].x;\n"
					"MOVR  R1.xy, R3;\n"
					"TXB   R1, R1.xyzz, texture[0], 2D;\n"
					"MADR  R1, R1, c[1].y, R0;\n"
					"MOVR  R0.z, c[1].x;\n"
					"MOVR  R0.xy, R2;\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MADR  R1, R0, c[1].y, R1;\n"
					"MOVR  R0.z, c[1].x;\n"
					"MOVR  R0.xy, R3.zwzw;\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MADR  R1, R0, c[1].y, R1;\n"
					"MOVR  R0.z, c[1].x;\n"
					"MOVR  R0.xy, R2.zwzw;\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MADR  R0, R0, c[1].y, R1;\n"
					"MOVR  R1.z, c[1].x;\n"
					"MOVR  R1.xy, R3.xwzw;\n"
					"TXB   R1, R1.xyzz, texture[0], 2D;\n"
					"MADR  R0, R1, c[1].z, R0;\n"
					"MOVR  R1.z, c[1].x;\n"
					"MOVR  R1.xy, R2.xwzw;\n"
					"TXB   R1, R1.xyzz, texture[0], 2D;\n"
					"MADR  R1, R1, c[1].z, R0;\n"
					"MOVR  R0.z, c[1].x;\n"
					"ADDR  R0.xy, fragment.texcoord[0], c[0].xzzw;\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MADR  R1, R0, c[1].z, R1;\n"
					"MOVR  R0.z, c[1].x;\n"
					"ADDR  R0.xy, fragment.texcoord[0], -c[0].xzzw;\n"
					"TXB   R0, R0.xyzz, texture[0], 2D;\n"
					"MADR  R0, R0, c[1].z, R1;\n"
					"MULR oCol, R0, c[1].z;\n"
					"END\n"
					"# 39 instructions, 4 R-regs, 0 H-regs\n"
					"\n", "v_pix:l:0", "n_tex");
				printf("\tusing low-level GL_ARB_fragment_program\n");
			}
			// try the same thing two times, once using assembler shader, then using GLSL shader

			if(!(p_shader_object = new CGLShaderObject(0, 0, &t_filter))) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			char *p_s_compile_log = 0;
			char *p_s_link_log = 0;
			if(!p_shader_object->Compile_Link(p_state, p_s_compile_log, p_s_link_log)) {
				fprintf(stderr, "error: error compiling shader:\n%s\n%s\n",
					(p_s_compile_log)? p_s_compile_log : "(null)",
					(p_s_link_log)? p_s_link_log : "(null)");
				if(p_s_compile_log)
					delete[] p_s_compile_log;
				if(p_s_link_log)
					delete[] p_s_link_log;
				return -1;
			}
			if(p_s_compile_log) {
				fprintf(stderr, "warning: while compiling shader:\n%s\n", p_s_compile_log);
				delete[] p_s_compile_log;
			}
			if(p_s_link_log) {
				fprintf(stderr, "warning: while linking shader:\n%s\n", p_s_link_log);
				delete[] p_s_link_log;
			}
			// compile and link shader

			if(!(p_shader_params = p_shader_object->p_GetParameterBuffer())) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			// get shader parameter buffer

			CShaderVectorParameterRef<float> t_pix_param = p_shader_params->t_Find_Parameter("v_pix");
			if(!t_pix_param.b_Valid())
				fprintf(stderr, "warning: v_pix uniform not found\n");
			else
				t_pix_param.SetValue(2.0f / 256, 2.0f / 256, -2.0f / 256, 0); // pixel_w, pixel_h, -pixel_h, 0
			// set pixel size
		}

		{
			if(!p_frame_buffer->b_Status() ||
			   !p_frame_buffer->Bind() ||
			   !p_frame_buffer->BindColorTexture_2D(*p_texture, GL_TEXTURE_2D)) {
				fprintf(stderr, "error: failed to bind FBO for r2t\n");
				return -1;
			}
			glViewport(0, 0, p_frame_buffer->n_Width(), p_frame_buffer->n_Height());
			// bind FBO for rendering to texture

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			// ortho

			p_state->ClearColor4f(1, 1, 1, 1);
			p_state->EnableDepthTest();
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			// clear color & depth

			p_shader_object->Bind(p_state, p_shader_params);
			// bind shader

			p_noise_texture->Bind_Enable(p_state);

			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			glVertex2f(-1, -1);
			glTexCoord2f(1, 0);
			glVertex2f( 1, -1);
			glTexCoord2f(1, 1);
			glVertex2f( 1,  1);
			glTexCoord2f(0, 1);
			glVertex2f(-1,  1);
			glEnd();
			// draw fullscreen quad

			if(!p_frame_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D, 0) ||
			   !p_frame_buffer->Release()) {
				fprintf(stderr, "error: failed to release FBO\n");
				return -1;
			}
			// release FBO

			p_state->BindProgramObject(0);
			// release shader
		}

		if(!SaveTexture((n_level)? "7_clock_box_glsl.tga" : "6_clock_box_arbfp.tga", p_texture)) {
			fprintf(stderr, "error: error saving texture to file\n");
			return -1;
		}
		// save texture to file

		delete p_shader_params;
		delete p_shader_object;
	}

	if(glGetError() != GL_NO_ERROR)
		fprintf(stderr, "error: OpenGL errors detected\n");
	// check

	printf("cleanup\n");

	delete p_shader;
	delete p_frame_buffer2;
	delete p_frame_buffer;
	delete p_noise_texture;
	delete p_texture;
	delete p_texture2;
	delete p_state;
	// cleanup

	return 0;
}
